/**
 * Copyright (C) 2009-2010 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <glib/gstdio.h>

#include <gtk/gtk.h>

#include "main.h"
#include "utils.h"
#include "settings.h"
#include "wk-view.h"
#include "wk-window.h"
#include "wk-browser.h"

#define DBUS_SIGNAL "com.linutop.WebKiosk.Signal"

struct _WkBrowserPrivate
{
    GtkWidget *main_window;
    GList *children; /* windows or tabs */
    GList *externals; /* list of pids to kill with the object */

    DBusConnection *bus;

    WkSettings *settings;
};

static void _browser_child_destroy_one (gpointer child, gpointer user_data);
static void _browser_external_kill_one (gpointer external, gpointer user_data);

static gboolean _browser_check_password (const gchar *crypted_passwd);

static DBusHandlerResult on_dbus_signal (DBusConnection *connection,
                                         DBusMessage *message,
                                         void *user_data);

static void wk_browser_class_init (WkBrowserClass *klass);
static void wk_browser_init (WkBrowser *browser);
static void wk_browser_finalize (GObject *object);

G_DEFINE_TYPE (WkBrowser, wk_browser, G_TYPE_OBJECT);

static void
wk_browser_class_init (WkBrowserClass *klass)
{
    g_type_class_add_private (klass, sizeof (WkBrowserPrivate));

    GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
    gobject_class->finalize = wk_browser_finalize;
}

static void
wk_browser_init (WkBrowser *browser)
{
    browser->priv = G_TYPE_INSTANCE_GET_PRIVATE (browser,
                                                WK_TYPE_BROWSER,
                                                WkBrowserPrivate);
}

static void
wk_browser_finalize (GObject *object)
{
    WkBrowser *browser = WK_BROWSER (object);

    wk_browser_cleanup (browser);

    g_object_unref (browser->priv->bus);
}

void
wk_browser_cleanup (WkBrowser *browser)
{
    gchar *tmp;

    gtk_widget_destroy (browser->priv->main_window);

    g_list_foreach (browser->priv->children, _browser_child_destroy_one, NULL);
    g_list_free (browser->priv->children);
    browser->priv->children = NULL;

    g_list_foreach (browser->priv->externals, _browser_external_kill_one, NULL);
    g_list_free (browser->priv->externals);
    browser->priv->externals = NULL;

    tmp = g_build_filename (g_get_tmp_dir (),
                            browser->priv->settings->tmpdir_name,
                            NULL);
    empty_dir (tmp);
    g_rmdir (tmp);
    g_free (tmp);

    stop_unclutter ();

    wk_settings_free (browser->priv->settings);
}

/* Implements the callbacks to signal received via dbus */
static DBusHandlerResult
on_dbus_signal (DBusConnection *connection, DBusMessage *message, void *user_data)
{
    WkBrowser *browser = WK_BROWSER (user_data);

    if (dbus_message_is_signal (message, DBUS_SIGNAL, "Quit"))
    {
        wk_browser_cleanup (browser);
        gtk_main_quit ();

        return DBUS_HANDLER_RESULT_HANDLED;
    }
    else if (dbus_message_is_signal (message, DBUS_SIGNAL, "StartSession"))
    {
        wk_browser_start_session (browser);

        return DBUS_HANDLER_RESULT_HANDLED;
    }
    else if (dbus_message_is_signal (message, DBUS_SIGNAL, "StopSession"))
    {
        wk_browser_clear_session (browser, TRUE);

        return DBUS_HANDLER_RESULT_HANDLED;
    }
    return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

GObject *
wk_browser_get (const gchar *uri)
{
    static WkBrowser *browser = NULL;
    WkSettings *settings;
    DBusError error;
    gchar *tmp;

    if (G_LIKELY (browser != NULL))
    {
        g_object_ref (G_OBJECT (browser));
        return G_OBJECT (browser);
    }


    /* creat the browser object */
    browser = g_object_new (WK_TYPE_BROWSER, NULL);
    g_object_add_weak_pointer (G_OBJECT (browser), (gpointer) &browser);

    /* FIXME: dynamic file name */
    wk_settings_load_from_uri (&browser->priv->settings,
                               uri ? uri
                                   : SYSCONFDIR "/xdg/webkiosk/preferences");

    settings = browser->priv->settings;
    tmp = g_build_filename (g_get_tmp_dir (), settings->tmpdir_name, NULL);
    g_mkdir_with_parents (tmp, 0700);
    g_free (tmp);
    tmp = NULL;

    browser->priv->main_window = wk_window_new (GTK_WINDOW_TOPLEVEL,
                                                "http://linutop.com");
    if (settings->fullscreen)
        gtk_window_fullscreen (GTK_WINDOW (browser->priv->main_window));

    gtk_widget_show_all (browser->priv->main_window);

    wk_view_start (WK_VIEW (wk_window_get_view (WK_WINDOW (browser->priv->main_window))));

    /* setup the dbus connection */
    dbus_error_init (&error);
    if ((browser->priv->bus = dbus_bus_get (DBUS_BUS_SYSTEM, &error)) == NULL)
    {
        g_warning ("Failed to connect to the D-BUS daemon: %s", error.message);
        dbus_error_free (&error);
        return NULL;
    }
    dbus_connection_setup_with_g_main (browser->priv->bus, NULL);
    dbus_bus_add_match (browser->priv->bus,
                        "type='signal',interface='com.linutop.WebKiosk.Signal'",
                        NULL);
    dbus_connection_add_filter (browser->priv->bus, on_dbus_signal,
                                (void *) browser, NULL);

    return (G_OBJECT (browser));
}

WkSettings *
wk_browser_get_settings (WkBrowser *browser)
{
    return browser->priv->settings;
}

static void
_browser_child_destroy_one (gpointer child, gpointer user_data)
{
    gtk_widget_destroy (GTK_WIDGET (child));
}

static void
_browser_external_kill_one (gpointer external, gpointer user_data)
{
    kill_external (GPOINTER_TO_INT (external));
}

static gboolean
_browser_check_password (const gchar *crypted_passwd)
{
    gboolean ret = FALSE;
    GtkWidget *d, *entry, *label;
    WkBrowser *browser;

    browser = WK_BROWSER (wk_browser_get (NULL));

    gdk_event_handler_set (event_no_filter, NULL, NULL);

    d = gtk_dialog_new_with_buttons (_("Administrator password"),
        GTK_WINDOW (browser->priv->main_window),
        GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
        GTK_STOCK_OK, GTK_RESPONSE_OK, NULL);

    label = gtk_label_new (_("Type your administrator password:"));
    gtk_widget_show (label);
    entry = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (entry), FALSE);
    gtk_entry_set_activates_default (GTK_ENTRY (entry), TRUE);
    gtk_widget_show (entry);

    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), label, FALSE, FALSE, 2);
    gtk_box_pack_start (GTK_BOX (GTK_DIALOG (d)->vbox), entry, FALSE, FALSE, 2);

    if (GTK_RESPONSE_OK == gtk_dialog_run (GTK_DIALOG (d)))
    {
        ret = validate_passwd (gtk_entry_get_text (GTK_ENTRY (entry)),
                               crypted_passwd);
    }

    gtk_entry_set_text (GTK_ENTRY (entry), "");
    gtk_widget_destroy (d);

    gdk_event_handler_set (event_filter, NULL, NULL);

    return ret;
}

void
wk_browser_clear_session (WkBrowser *browser, gboolean stop)
{
    GtkWidget *view;
    gchar *tmp;

    g_list_foreach (browser->priv->children, _browser_child_destroy_one, NULL);
    g_list_free (browser->priv->children);
    browser->priv->children = NULL;

    g_list_foreach (browser->priv->externals, _browser_external_kill_one, NULL);
    g_list_free (browser->priv->externals);
    browser->priv->externals = NULL;

    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    if (stop)
        wk_view_stop_session (WK_VIEW (view));
    else
        wk_view_clear_history (WK_VIEW (view));

    tmp = g_build_filename (g_get_tmp_dir (),
                            browser->priv->settings->tmpdir_name,
                            NULL);
    empty_dir (tmp);
    g_free (tmp);
}

void
wk_browser_show_item (WkBrowser *browser, guint nth)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_select_item_nth (WK_VIEW (view), nth);
}

void
wk_browser_show_next (WkBrowser *browser)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_select_item_next (WK_VIEW (view), TRUE, FALSE);
}

void
wk_browser_show_prev (WkBrowser *browser)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_select_item_prev (WK_VIEW (view), TRUE, FALSE);
}

void
wk_browser_get_playlist (WkBrowser *browser, gchar **playlist)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_get_playlist_as_str (WK_VIEW (view), playlist);
}

void
wk_browser_add_time (WkBrowser *browser, guint seconds)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_add_time (WK_VIEW (view), seconds);
}

void
wk_browser_start_session (WkBrowser *browser)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_start_session (WK_VIEW (view));
}

void
wk_browser_start_slideshow (WkBrowser *browser)
{
    GtkWidget *view;
    view = wk_window_get_view (WK_WINDOW (browser->priv->main_window));
    wk_view_start_slideshow (WK_VIEW (view));
}

gint
list_compare_pid (gconstpointer a, gconstpointer b)
{
    return GPOINTER_TO_INT (a) - GPOINTER_TO_INT (b);
}

void
wk_browser_register_external (WkBrowser *browser, GPid pid)
{
    if (!g_list_find_custom (browser->priv->externals,
                             GINT_TO_POINTER (pid),
                             list_compare_pid))
    {
        browser->priv->externals = g_list_append (browser->priv->externals,
                                                  GINT_TO_POINTER (pid));
    }
}

void
wk_browser_unregister_external (WkBrowser *browser, GPid pid, gboolean with_kill)
{
    if (g_list_find_custom (browser->priv->externals,
                            GINT_TO_POINTER (pid),
                            list_compare_pid))
    {
        browser->priv->externals = g_list_remove (browser->priv->externals,
                                                  GINT_TO_POINTER (pid));
        if (with_kill)
            kill_external (pid);
    }
}

void
wk_browser_register_window (WkBrowser *browser,
                            WkWindow  *window)
{
    browser->priv->children = g_list_append (browser->priv->children, window);
}

void
wk_browser_unregister_window (WkBrowser *browser,
                              WkWindow  *window)
{
    browser->priv->children = g_list_remove (browser->priv->children, window);
}

gboolean
wk_browser_maybe_quit (WkBrowser *browser)
{
    gboolean do_quit = TRUE;
    WkSettings *settings = browser->priv->settings;

    if (!settings->can_quit)
    {
        if (settings->admin_passwd)
            do_quit = _browser_check_password (settings->admin_passwd);
        else
            do_quit = FALSE;
    }

    if (do_quit)
    {
        wk_browser_cleanup (browser);
        gtk_main_quit ();
    }

    return do_quit;
}

gboolean
wk_browser_has_child (WkBrowser *browser)
{
    return g_list_length (browser->priv->children) > 0;
}

GtkWidget *
wk_browser_get_main_window (WkBrowser *browser)
{
    return browser->priv->main_window;
}

void
wk_browser_promote_first_child (WkBrowser *browser)
{
    GList *l;

    gtk_widget_destroy (browser->priv->main_window);
    l = g_list_first (browser->priv->children);
    browser->priv->main_window = GTK_WIDGET (l->data);

    browser->priv->children = g_list_next (l);
}

guint
wk_browser_get_windows_count (WkBrowser *browser)
{
    return 1 + g_list_length (browser->priv->children);
}
