/**
 * Copyright (C) 2009-2010 Gauvain Pocentek <gpocentek@linutop.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_PWD_H
#include <pwd.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include "settings.h"
#include "gtk-extensions.h"

#include "xfce_framebox.h"
#include "config-main.h"

#include "main.h"
#include "utils.h"

static GHashTable *hash;
static GtkWidget *kiosk_btn, *display_btn;
static GtkWidget *at_startup;
static GtkWidget *use_pwd, *pwd_table;
static GtkWidget *pwd_entry1, *pwd_entry2, *pwd_btn;
gboolean was_at_startup;

extern GtkWidget *kiosk_widget;
extern GtkWidget *display_widget;

ConfigItemDesc main_descs[] =
{
    { "main", "mode", CONFIG_ITEM_TYPE_STRING, "kiosk" },
    { "main", "admin_passwd", CONFIG_ITEM_TYPE_STRING, "" },
    { "main", "fullscreen", CONFIG_ITEM_TYPE_BOOLEAN, "true" },
    { NULL }
};

struct shortcut
{
    const gchar *shortcut;
    const gchar *detail;
} shortcuts[] =
{
    { "ctrl+q, Esc", N_("Leave the application") },
    { "ctrl+w", N_("Close the active window") },
    { "ctrl+l", N_("Get the focus on the address bar") },
    { NULL, NULL }
};

inline GHashTable *
config_main_get_hash (WkKeyFile *kf)
{
    hash = handled_settings_get_from_array (kf, main_descs);
    return hash;
}

static gboolean
get_is_active ()
{
    uid_t uid = 0;
    struct passwd *pw;
    const gchar *cuid = NULL;
    gchar *out = NULL;
    gchar *err = NULL;
    gchar *cmd = NULL;
    gchar *cfile = NULL;
    gboolean ret = FALSE;

    cuid = g_getenv ("SUDO_UID");
    if (cuid)
    {
        uid = atoi (cuid);
        pw = getpwuid (uid);
        cfile = g_strdup_printf ("%s/.config/appstarterrc", pw->pw_dir);
    }
    else
        cfile = g_strdup_printf ("%s/.config/appstarterrc", g_getenv ("HOME"));

    cmd = g_strdup_printf ("appstarter --selected -c %s", cfile);
    g_spawn_command_line_sync (cmd, &out, &err, NULL, NULL);

    if (strncmp ("webkiosk", out, 8) == 0)
        ret = TRUE;

    g_free (cfile);
    g_free (cmd);
    g_free (out);
    g_free (err);

    was_at_startup = ret;

    return ret;
}

static void
set_is_active ()
{
    uid_t uid = 0;
    struct passwd *pw;
    const gchar *cuid = NULL;
    gchar *cmd = NULL;
    gchar *cfile = NULL;

    if (gtk_toggle_button_get_active (
            GTK_TOGGLE_BUTTON (at_startup)) == was_at_startup)
    {
        /* nothing changed, do nothing */
        return;
    }

    cuid = g_getenv ("SUDO_UID");
    if (cuid)
    {
        uid = atoi (cuid);
        pw = getpwuid (uid);
        cfile = g_strdup_printf ("%s/.config/appstarterrc", pw->pw_dir);
    }
    else
        cfile = g_strdup_printf ("%s/.config/appstarterrc", g_getenv ("HOME"));

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (at_startup))
        && !was_at_startup)
    {
        cmd = g_strdup_printf ("appstarter --select 'webkiosk' "
                               "--command 'linutop-webkiosk' "
                               "-c %s", cfile);
    }
    else if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (at_startup))
        && was_at_startup)
    {
        cmd = g_strdup_printf ("appstarter --select 'none' "
                               "--command 'true' "
                               "-c %s", cfile);
    }

    g_spawn_command_line_sync (cmd, NULL, NULL, NULL, NULL);
    if (cuid)
        chown (cfile, uid, -1);

    g_free (cfile);
    g_free (cmd);
}


static void
on_btn_toggled (GtkToggleButton *btn, gpointer data)
{
    GtkNotebook *notebook = GTK_NOTEBOOK (data);
    ConfigItem *item;

    item = g_hash_table_lookup (hash, "main:mode");

    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (kiosk_btn)))
    {
        wk_config_item_set_value (item, (gpointer) g_strdup ("kiosk"), TRUE);

        if (gtk_notebook_get_n_pages (notebook) == NOTEBOOK_PAGE_NB)
            return;

        gtk_notebook_insert_page (notebook,
                                  kiosk_widget,
                                  gtk_label_new (_("Web Kiosk")),
                                  NOTEBOOK_PAGE_KIOSK);
        /* drop the reference we created to keep the widget in memory */
        g_object_unref (kiosk_widget);
    }
    else
    {
        wk_config_item_set_value (item, (gpointer) g_strdup ("display"), TRUE);

        if (gtk_notebook_get_n_pages (notebook) != NOTEBOOK_PAGE_NB)
            return;

        /* add a reference to keep the widget around */
        g_object_ref (kiosk_widget);
        gtk_notebook_remove_page (notebook, NOTEBOOK_PAGE_KIOSK);
    }
}

static void
on_pwd_entry_changed (GtkEntry *entry, gpointer data)
{
    const gchar *txt1 = gtk_entry_get_text (GTK_ENTRY (pwd_entry1));
    const gchar *txt2 = gtk_entry_get_text (GTK_ENTRY (pwd_entry2));
    if (txt1 && (g_strcmp0 (txt1, txt2) == 0))
        gtk_widget_set_sensitive (pwd_btn, TRUE);
    else
        gtk_widget_set_sensitive (pwd_btn, FALSE);
}

static void
on_pwd_btn_clicked (GtkButton *btn, gpointer data)
{
    gchar *crypted_passwd = NULL;
    ConfigItem *item;

    crypted_passwd = crypt_passwd (
        gtk_entry_get_text (GTK_ENTRY (pwd_entry1)), NULL);

    item = g_hash_table_lookup (hash, "main:admin_passwd");
    wk_config_item_set_value (item, crypted_passwd, TRUE);

    gtk_entry_set_text (GTK_ENTRY (pwd_entry1), "");
    gtk_entry_set_text (GTK_ENTRY (pwd_entry2), "");
}

static void
on_use_pwd_toggled (GtkToggleButton *btn, gpointer data)
{
    ConfigItem *item;

    if (gtk_toggle_button_get_active (btn))
        gtk_widget_set_sensitive (pwd_table, TRUE);
    else
    {
        item = g_hash_table_lookup (hash, "main:admin_passwd");
        wk_config_item_set_value (item, g_strdup (""), TRUE);
        gtk_widget_set_sensitive (pwd_table, FALSE);
    }
}

GtkWidget *
config_main_get_widget (GtkWidget *notebook)
{
    GtkWidget *mvbox;
    GtkWidget *vbox, *table, *hbox;
    GtkWidget *label, *img, *frame;
    gchar *markup = NULL;
    gchar *mode = NULL;
    gchar *pwd = NULL;
    int i;

    mvbox = gtk_vbox_new (FALSE, 2);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (mvbox), hbox, FALSE, FALSE, 0);

    frame = xfce_framebox_new (_("Main use"), TRUE);
    table = gtk_table_new (2, 4, FALSE);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), table);
    gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, FALSE, 6);

    kiosk_btn = gtk_radio_button_new (NULL);
    gtk_table_attach (GTK_TABLE (table), kiosk_btn, 0, 1, 0, 1,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);
    display_btn = gtk_radio_button_new_from_widget (
        GTK_RADIO_BUTTON (kiosk_btn));
    gtk_table_attach (GTK_TABLE (table), display_btn, 0, 1, 2, 3,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);

    label = gtk_label_new (_("Web Kiosk"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_table_attach (GTK_TABLE (table), label, 1, 2, 0, 1,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);
    label = gtk_label_new (_("Display"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    gtk_table_attach (GTK_TABLE (table), label, 1, 2, 2, 3,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);

    label = gtk_label_new (NULL);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    markup = g_strdup_printf ("<span style=\"italic\">%s</span>",
                              _("Secured Internet access"));
    gtk_label_set_markup (GTK_LABEL (label), markup);
    g_free (markup);
    gtk_table_attach (GTK_TABLE (table), label, 1, 2, 1, 2,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);

    label = gtk_label_new (NULL);
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    markup = g_strdup_printf ("<span style=\"italic\">%s</span>",
                              _("Digital display of web pages, images and videos"));
    gtk_label_set_markup (GTK_LABEL (label), markup);
    g_free (markup);
    gtk_table_attach (GTK_TABLE (table), label, 1, 2, 3, 4,
                      GTK_SHRINK | GTK_FILL, 0, 0, 0);

    g_signal_connect (G_OBJECT (kiosk_btn), "toggled",
                      G_CALLBACK (on_btn_toggled), notebook);
    g_signal_connect (G_OBJECT (display_btn), "toggled",
                      G_CALLBACK (on_btn_toggled), notebook);

    mode = (gchar *) wk_config_item_get_value (
                        g_hash_table_lookup (hash, "main:mode"));

    if (g_strcmp0 (mode, "kiosk") == 0)
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (kiosk_btn), TRUE);
    else
        gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (display_btn), TRUE);

    hbox = gtk_hbox_new (FALSE, 0);
    gtk_box_pack_start (GTK_BOX (mvbox), hbox, FALSE, FALSE, 0);

    vbox = gtk_vbox_new (FALSE, 2);

    gtk_box_pack_start (GTK_BOX (vbox),
                        wk_gtk_check_button_new (_("Fullscreen"),
                            g_hash_table_lookup (hash, "main:fullscreen"),
                            FALSE),
                        FALSE, FALSE, 0);

    at_startup = gtk_check_button_new_with_label (_("Run the kiosk at startup"));
    gtk_box_pack_start (GTK_BOX (vbox), at_startup, FALSE, FALSE, 0);
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (at_startup), get_is_active ());

    use_pwd = gtk_check_button_new_with_label (_("Set a password for admin tasks"));
    gtk_box_pack_start (GTK_BOX (vbox), use_pwd, FALSE, FALSE, 0);

    pwd_table = gtk_table_new (2, 3, FALSE);
    gtk_box_pack_start (GTK_BOX (vbox), pwd_table, FALSE, FALSE, 0);

    label = gtk_label_new (_("Password:"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    pwd_entry1 = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (pwd_entry1), FALSE);
    gtk_table_attach (GTK_TABLE (pwd_table), label, 0, 1, 0, 1,
                      GTK_FILL, GTK_FILL, 0, 0);
    gtk_table_attach (GTK_TABLE (pwd_table), pwd_entry1, 1, 2, 0, 1,
                      GTK_FILL, GTK_FILL, 0, 0);

    label = gtk_label_new (_("Retype it:"));
    gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
    pwd_entry2 = gtk_entry_new ();
    gtk_entry_set_visibility (GTK_ENTRY (pwd_entry2), FALSE);
    gtk_table_attach (GTK_TABLE (pwd_table), label, 0, 1, 1, 2,
                      GTK_FILL, GTK_FILL, 0, 0);
    gtk_table_attach (GTK_TABLE (pwd_table), pwd_entry2, 1, 2, 1, 2,
                      GTK_FILL, GTK_FILL, 0, 0);

    pwd_btn = gtk_button_new ();
    img = gtk_image_new_from_stock (GTK_STOCK_OK, GTK_ICON_SIZE_BUTTON);
    gtk_button_set_image (GTK_BUTTON (pwd_btn), img);
    gtk_button_set_label (GTK_BUTTON (pwd_btn), "Change");
    gtk_widget_set_sensitive (pwd_btn, FALSE);
    gtk_table_attach (GTK_TABLE (pwd_table), pwd_btn, 2, 3, 1, 2,
                      GTK_FILL, GTK_FILL, 0, 0);

    frame = xfce_framebox_new (_("Misc"), TRUE);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), vbox);
    gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, FALSE, 6);

    g_signal_connect (G_OBJECT (pwd_entry1), "changed",
                      G_CALLBACK (on_pwd_entry_changed), NULL);
    g_signal_connect (G_OBJECT (pwd_entry2), "changed",
                      G_CALLBACK (on_pwd_entry_changed), NULL);
    g_signal_connect (G_OBJECT (pwd_btn), "clicked",
                      G_CALLBACK (on_pwd_btn_clicked), NULL);


    g_signal_connect (G_OBJECT (use_pwd), "toggled",
                      G_CALLBACK (on_use_pwd_toggled), NULL);
    pwd = (gchar *) wk_config_item_get_value (
                        g_hash_table_lookup (hash, "main:admin_passwd"));
    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (use_pwd),
        (g_strcmp0 (pwd, "") != 0) ? TRUE : FALSE);

    table = gtk_table_new (2, 3, FALSE);
    gtk_table_set_col_spacings (GTK_TABLE (table), 8);

    for (i=0; shortcuts[i].shortcut; i++)
    {
        label = gtk_label_new ("");
        gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
        markup = g_strdup_printf ("<span weight=\"bold\">%s</span>",
                                  shortcuts[i].shortcut);
        gtk_label_set_markup (GTK_LABEL (label), markup);
        g_free (markup);
        gtk_table_attach (GTK_TABLE (table), label, 0, 1, i, i+1,
                          GTK_FILL, GTK_FILL, 0, 0);

        label = gtk_label_new (_(shortcuts[i].detail));
        gtk_misc_set_alignment (GTK_MISC (label), 0, 0.5);
        gtk_table_attach (GTK_TABLE (table), label, 1, 2, i, i+1,
                          GTK_FILL, GTK_FILL, 0, 0);
    }

    frame = xfce_framebox_new (_("Keyboard shortcuts"), TRUE);
    xfce_framebox_add (XFCE_FRAMEBOX (frame), table);
    gtk_box_pack_start (GTK_BOX (hbox), frame, FALSE, FALSE, 6);

    return mvbox;
}

void
config_main_save_config (WkKeyFile *kf, GHashTable *hash)
{
    ConfigItem *item;

    handled_settings_set_from_hash (kf, hash);

    /* override the kiosk config if we use the soft for display */
    if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (display_btn)))
    {
        item = wk_key_file_get_config_item (kf, "main", "start_with_session",
                                            CONFIG_ITEM_TYPE_BOOLEAN, FALSE);
        wk_config_item_set_value (item, GINT_TO_POINTER (0), FALSE);
        wk_key_file_set_config_item (kf, item);
    }

    set_is_active ();
}
